from enum import Enum
from collections import deque, defaultdict, Counter
import time
from numbers import Number

import numpy as np

"""A global counter to count the total number of VarAutoName instances"""
G_VAR_AUTO_NAME_NUM = 0


class Mode(Enum):
    Forward = 0
    Reverse = 1
    Mix = 2
    NONE = -1


class Var(object):
    """
        A class that holds variables with AutoDiff method in both forward and reverse mode.
        Supports 20+ elementary functions (see ../ops.py) for scalar(Number) and 2D matrix (np.ndarray).

        Attributes:
            val (Number or np.ndarray): numerical value of this variable.
            name (str or Number, optional): name assigned to the variable, such as "x","longlongname" or 1.1, 2.
            derivative (dict): a dict that stores the derivative values w.r.t all involved variables. For example,
                {"x": numerical value of dfdx, "y": numerical value of dfdy}. A 4D matrix D is used to represent
                the single der such as dcdx: D_{ijkl} represents dc_{ij}/dx_{kl}.
            mode (Mode): for newly declared instance, specify Forward or Reverse mode. The Mix mode will come from
                the operation of a Forward mode and a non-Forward mode.
            _var_shapes (dict): a dict that stores the shape of the involved variables used for squeezing the 4D
                derivative matrix to 2D Jacobian matrix if necessary. For example, {"x": (1,1), "y": (2,1)}.
                LEAVE IT ALONE when declare a new Var instance.
            _context (list): [cls, operands] where "cls" represents the operation and operands is [a, b] if cls is
                a binary operation else [a] for unary operation. LEAVE IT ALONE when declare a new Var instance.

            _bpgrad (dict): a dict stores the temporary backpropagation gradiant, used for reverse mode only.
                For example, u.bp = {f: 1} means dfdu = 1. The key here is the hash value of f (output, a Var instance)
                while the key in u.derivative is the name of x (input, a Var instance). Similarly, a 4D matrix D is used
                to represent a single gradient dfdc: D_{ijkl} represents df_{ij}/dc_{kl}.
            _degree (int): "out degree" = number of usage in the computational graph, used for reverse mode only.
        """

    def __init__(self, val, name=None, derivative=None, mode=Mode.Forward, _var_shapes=None, _context=None):
        """
        ----------
        Args:
            val (Number or np.ndarray): numerical value of the variable.
            name (str or Number, optional): name of the variable. If None(default), the variable will be treated
                as a constant, which means no derivative wrt this instance.
            derivative (Number or np.ndarray, optional): a dict; Defaults to None. If name is None, derivative will be
                set as an empty dict; If name is not None, derivative will be initialized as {name: 4D_matrix};
                Number/np.ndarray can be passed in as the `seed` for this variable (name should not be None and the
                shape of seed should match its value).
            mode (Mode): Forward(default)/Reverse. The mode of const will be set as Mode.NONE.

            _var_shapes (dict or None): Leave it None when declare an instance. See explanations above.
            _context (list or None): Leave it None when declare an instance. See explanations above.

        TO use:
            >>> x = Var(1, 'x', mode=Mode.Forward)
            >>> y = Var(np.array([[1],[2]]), 'y', mode=Mode.Reverse)
        """

        self._val = val

        if name is None or isinstance(name, (str, Number)):
            self.name = name
        else:
            raise TypeError(f"name should be a str or Number, {type(name)} is not supported.")

        # Init derivative
        if isinstance(derivative, dict):
            self.derivative = derivative
        elif name is not None:
            self.derivative = {name: self._init_seed(val, derivative)}
        else:
            if derivative is not None:
                raise ValueError(f"Need a name!")
            # Use {} instead of None to skip the type check when self.derivative is used
            self.derivative = {}

        # Be careful, this dict is designed for sharing across multiple instances
        # which means for x = Var(1, 'x'), x.self._var_shapes can contain key="y" that is not x's "target wrt var"
        self._var_shapes = _var_shapes
        if name is not None:
            try:
                self._var_shapes[name] = np.shape(val)
            except:
                self._var_shapes = {name: np.shape(val)}

        self.mode = Mode.NONE if self.is_const else mode
        self._context = _context

        # Used only for reverse mode
        # Will be activated when ._reverse_diff() is called
        self._degrees = None
        self._bpgrad = None

    def _init_seed(self, val, seed=None):
        """
        Initialize the derivative for newly declared var instance. The shape of seed should match the shape of val.
        Or exception will be thrown out. If val is scalar, seed must be a scalar too; If val is matrix, seed could
        be a scalar or a matrix.
        Args:
            val: var's value, used for aligning the shape of val and derivative
            seed: a Number or np.ndarray, defaults to None.

        Returns:
            a 4D matrix as the initial derivative.

        For example (this function will be called in __init__):
        >>> Var(1, 'x', 100).derivative['x'].tolist()
        [[[[100.0]]]]
        >>> Var(np.array([[1],[2]]), 'x', 2).derivative['x'].tolist() # output is np.ndarray
        [[[[2.0], [0.0]]], [[[0.0], [2.0]]]]
        >>> Var(np.array([[1],[2]]), 'x', np.array([[100],[200]])).derivative['x'].tolist()
        [[[[100.0], [0.0]]], [[[0.0], [200.0]]]]
        """
        if seed is None:
            seed = 1
        elif not isinstance(seed, (Number, np.ndarray, list)):
            raise TypeError(f"Init derivative(seed) should be a ndarray or Number, {type(seed)} is not supported.")

        seed = np.array(seed)

        ndim = np.ndim(val)
        # Init seed should be a scalar or the shape of seed should be equal to the shape of  value
        assert np.ndim(seed) == 0 or np.size(seed) == np.size(val), (
            f"Initial derivative {np.shape(seed)} should match the shape of val {np.shape(val)}")

        if ndim == 2:
            k, l = val.shape
        elif ndim == 0:
            k = l = 1
        else:
            raise ValueError(f"Val only support scalar/2D-matrix. Input: {val.shape}")

        return np.einsum('ij,kl->ikjl', np.eye(k) * seed, np.eye(l))

    def __str__(self):
        return f"(val: {self.val}, der: {self.derivative})"

    def __repr__(self):
        return f"({self.__class__} name: {self.name} val: {self.val}, der: {self.derivative})"

    def __eq__(self, b):
        """Only compare the `val` and `derivative`. `name` is ignored."""
        if not isinstance(b, Var) or not np.allclose(self.val, b.val) or not (
                self.derivative.keys() == b.derivative.keys()):
            return False

        for wrt in self.derivative.keys():
            # TODO: Fan
            # Use np.array_equal instead to check the shape?
            if not np.allclose(self.derivative[wrt], b.derivative[wrt]):
                return False

        return True

    @property
    def val(self):
        """Return numerical value of variable"""
        return self._val

    @val.setter
    def val(self, v):
        """Set numerical value for variable"""
        self._val = v

    def _squeeze_der(self, name, v):
        """
        Squeeze the 4D derivative matrix to match the expectation of Jacobian matrix. The output shape is listed below:
        Input type --> output type: Jacobian matrix type
        Scalar --> scalar: scalar
        Scalar --> vector((n,1) or (1,n)): 2D matrix(n,1)
        Vector((n,1) or (1,n)) --> scalar: 2D matrix(1,n)
        Vector((n,1) or (1,n)) --> Vector((m,1) or (1,m)): 2D matrix(m,n)
        Matrix((m,n)) --> matrix((p,q)): 3D matrix if one of m,n,p,q is 1 else 4D matrix
        Args:
            name: name of target var instance
            v: 4D derivative matrix

        Returns:
            A scalar or matrix, the squeezed derivative.
        """
        shape = self._var_shapes[name]

        if len(shape) == 0:
            try:
                return v.item()
            except:
                return np.squeeze(np.squeeze(v, -1), -1)

        m, n, k, l = v.shape
        assert (k, l) == shape, f"var shape {shape} and der shape: {self.val.shape} mismatch!"

        if l == 1:
            v = np.squeeze(v, -1)
        elif k == 1:
            v = np.squeeze(v, -2)

        if n == 1:
            v = np.squeeze(v, 1)
        elif m == 1:
            v = np.squeeze(v, 0)

        return v

    def __hash__(self):
        return id(self)

    def _count_degrees(self):
        """
        Count "out degree" for every involved var instance for reverse mode.
        Returns: a dict where key = node, val = out degree
        """

        q = deque()
        q.append(self)

        degrees = Counter()
        visited = defaultdict(bool)

        while len(q) > 0:
            v = q.popleft()
            if v._context is None:
                continue
            _, operands = v._context
            for t in operands:
                degrees[t] += 1
                if not visited[t]:
                    visited[t] = True
                    q.append(t)

        return degrees

    def _backward(self):
        """
        Recursively trace back along the computational graph to propagate the derivative from output to input.
        See more explanations in code comments.
        """

        # Two cases to "merge" the .derivative from the forward propagation and ._bpgrad from the back propagation
        # if self.derivative is not None, two possible cases:
        # 1. For target vars like x,y whose .derivative is initialized when declared;
        # 2. For mix mode calculation, some node in forward mode in the trace has non-empty .derivative
        # Be careful, the merged derivative could be part of the total derivative so we need to accumulate all.
        if len(self.derivative) > 0:
            from pyautodiff import Ops
            # f: a var instance, dfdc: numerical derivative for dfdc (suppose current instance(self) is c)
            f, dfdc = self._bpgrad.popitem()
            # Merge two 4D matrix
            d = Ops.merge_fwd_backprop(self.derivative, dfdc)

            # Accumulate the derivatives
            f.derivative = Counter(f.derivative)
            f.derivative.update(d)
            f.derivative = dict(f.derivative)

        elif self._context is not None:
            cls, operands = self._context
            cls.backprop(operands[0],
                         operands[1] if len(operands) == 2 else None,
                         self.val,
                         self._bpgrad)
            # Clear it for next BP
            self._bpgrad.popitem()

            for t in operands:
                t._degree -= 1
                # When t.degree is 0, dfdt is complete and safe to trace back
                if t._degree == 0:
                    t._backward()

    def _reverse_diff(self):
        """
        Start AD of reverse mode.
        """
        degrees = self._count_degrees()
        for t in degrees:
            t._degree = degrees[t]
            t._bpgrad = Counter()

        self._bpgrad = {self: self._init_seed(self.val)}
        self._backward()

    def diff(self, wrts=None, return_raw=False):
        """
        Get derivative w.r.t. to each var in `wrts`.
        Args:
            wrts: single variable name or a list/tuple of variable names. Defaults to None, equals to `all`.

        Returns:
            a Number or np.ndarray if wrts is single variable name;
            or a list of Number or np.ndarray that corresponds to each variable name in wrts, if wrts is a list/tuple;
            or a dict with the variable name as a key and value as a Number or np.ndarray, if wrts is None.

        """

        # Reverse mode
        if len(self.derivative) == 0 and self._context is not None:
            self._reverse_diff()

        der = self.derivative
        keys = list(der.keys())

        if not return_raw:
            der = {x: self._squeeze_der(x, der[x]) for x in keys}

        if wrts is None:
            if len(keys) == 0:
                return 0

            if len(keys) == 1:
                return der[keys[0]]

            return der
        elif isinstance(wrts, (list, tuple)):
            return [der.get(w, 0) for w in wrts]
        else:
            try:
                return der[wrts]
            except:
                raise TypeError(f"wrts only supports None/list/tuple or a var name!")

    @property
    def T(self):
        """
        To support x.T
        Returns: Transposed matrix

        """
        return self.transpose()

    @property
    def is_const(self):
        """Const like: Var(1)"""
        return self._var_shapes is None


class VarAutoName(Var):
    """
    A wrapper class for class `Var`. Variable names are auto-generated by combining the current number of
    instances of `VarAutoName` and system process time to avoid duplicate names.
    """

    def __init__(self, val, derivative=None, mode=Mode.Forward):
        """

        Args:
            val (Number or np.ndarray): numerical value; same as `val` in `Var`.
            derivative: a dict or a Number/np.ndarray. Defaults to None. Same as `derivative` in `Var`.
        """
        # TODO: Fan
        # Add a Lock to protect G_VAR_AUTO_NAME_NUM
        global G_VAR_AUTO_NAME_NUM
        G_VAR_AUTO_NAME_NUM += 1
        name = f"{G_VAR_AUTO_NAME_NUM}_{time.process_time()}"

        super().__init__(val, name=name, derivative=derivative, mode=mode)

    @staticmethod
    def clear_var_counter():
        """
        Clears G_VAR_AUTO_NAME_NUM in case of overflow.

        Returns: None

        """
        global G_VAR_AUTO_NAME_NUM
        G_VAR_AUTO_NAME_NUM = 0



import math
from collections import defaultdict
from functools import wraps

import numpy as np

from pyautodiff import Var, Mode


def _dunder_wrapper(fn, is_unary=False):
    """
    A wrapper function to bridge dunder method and classmethod (operation).
    For example, Var.__add__ = dunder_wrapper(VarAdd.binary_operation).
    Or Var.__add__ = lambda a, b: VarAdd.binary_operation(a, b)
    Args:
        fn: operation function
        is_unary: Defaults to False for binary operations like Add, Substract; True for unary operations like abs, exp.

    Returns:
        The wrapped function.
    """

    @wraps(fn)
    def wrapper(*args):
        a = args[0]
        if is_unary:
            return fn(a)

        b = args[1]
        return fn(a, b)

    return wrapper


class Ops:
    """
    A template class for all operations for class `Var` (i.e. unary and binary operations).
    For each operation, users MUST implement (at least) two functions: `op()` and `local_derivative()`.
    Non-element-wise operations should re-write some more methods. See 'VarTranspose' and 'VarMatMul' as reference.
    Then the propagation for forward/reverse/mix mode will be auto handled by the pipeline, which locate in
    `binary_operation()` and `unary_operation()`.
    """

    # Unary operation: 1 (by default)
    # Binary operaton: 2 (customized)
    n_operands = 1

    # A string to be displayed in computational graph.
    # If None, suffix of class name will be used.
    # For example, exp (operation) will show `Exp` since its class name is `VarExp` in the plot.
    # See ../visualization.py for its usage.
    symbol = None

    @classmethod
    def op(cls, va, vb):
        """
        Implement the numerical value operation in this function.
        For unary operation: vc = f(va), return vc;
        For binary operation: vc = va op vb, return vc;
        To be implemented by each operation.
        Args:
            va: numerical value of operand a (a.val, a is a Var instance);
            vb: numerical value of operand b (b.val, b is a Var instance);

        Returns:
            A Number or np.ndarray, the numerical value of this operation
        """

        raise NotImplementedError

    @classmethod
    def local_derivative(cls, va, vb, vc, skip_lda=False, skip_ldb=False):
        """
        Calculate the derivative for every elementary operations.
        For unary operation: c = f(a), return the local partial derivative: df/da;
        For binary operation: c = a op b, return the local partial derivatives df/da, df/db;
        (a,b could be results of some operations)
        For example,

            x = Var(1, 'x')
            a = x + 1
            b = 2 - x
            c = a * b

        The local derivative dc/da = 1, dc/db = 2;
        The target derivative dc/dx = dc/da*da/dx + dc/db*db/dx = 1*1 + 2*(-1) = -1

        Args:
            va: numerical value of operand (a Var instance) a (=a.val);
            vb: numerical value of operand (a Var instance) b (=b.val);
            vc: numerical value of operation result (a Var instance) c (=c.val);
            skip_lda: If a is a constant, no need to calculate dc/da
            skip_ldb: If b is a constant, no need to calculate dc/db
        Returns:
            A Number or np.ndarray for unary operation;
            A list of two Numbers or np.ndarrays for binary operation;

        """
        raise NotImplementedError

    @classmethod
    def chain_rule(cls, lda, da, ldb, db, forward_mode=True):
        """
        Apply chain rule in forward mode.
        For composite function: c = g(f(a, b)), dg/dx = dg/df*df/dx; dg/dy = dg/df*df/dy;
        Args:
            lda: A Number or np.ndarray, represents the local derivative: dc/da
            da: a dict stores the derivative of a.
                For example, {'x': da/dx, 'y': da/dy}, where `x`,'y' is the involved variables.
            ldb: A Number or np.ndarray, represents the local derivative: dc/db
            db: a dict stores the derivative of b.
                For example, {'x': db/dx, 'y': db/dy}, where `x`,'y' is the involved variables.
            forward_mode: defaults to True; False for reverse or mix mode.
        Returns:
            A dict stores the derivative of c by applying the chain rule. For example,
            {'x': dc/dx, 'y': dc/dy} where `x`,'y' are the target variables.

        """

        einsum_dispatcher = "ijkl,ij->ijkl" if forward_mode else "ijkl,kl->ijkl"

        def _apply(d, ld):
            if d is None:
                return

            ndim = np.ndim(ld)
            if ndim == 0:
                fn = lambda tot, loc: tot * loc
            elif ndim == 2:
                fn = lambda tot, loc: np.einsum(einsum_dispatcher, tot, loc)
            else:
                raise TypeError(f"Local derivative only supports scalar or 2D matrix but not {np.shape(ld)}")

            for wrt in d:
                dc[wrt] += fn(d[wrt], ld)

        dc = defaultdict(int)
        _apply(da, lda)
        _apply(db, ldb)

        return dict(dc)

    @classmethod
    def merge_var_shapes(cls, sa, sb=None):
        """
        Propagate the _var_shapes to the operation result by synthesizing the _var_shapes of a and b.
        BE CAREFUL, a _var_shapes (dict) instance can be shared across multiple var instances.
        Don't use _var_shapes for any instance specific calculation.
        Args:
            sa: _var_shapes of the first operand
            sb: _var_shapes of the second operand, could be None

        Returns:
            a dict, the merged _var_shapes
        """
        if sb is None:
            return sa

        if sa is None:
            return sb

        sa.update(sb)
        return sa

    @classmethod
    def merge_modes(cls, ma, mb=None):
        """
        Merge mode by such rules:
        1. Forward op reverse/mix --> mix
        2. Forward op forward/NONE --> forward
        3. Reverse op reverse/NONE --> reverse
        4. Reverse/mix/NONE op mix --> mix
        5. NONE op NONE --> NONE
        Args:
            ma: a.mode
            mb: b.mode

        Returns:
            A mode value

        """
        if mb is None or mb == Mode.NONE:
            return ma

        if ma == Mode.NONE:
            return mb

        if ma != mb:
            return Mode.Mix

        return ma

    @classmethod
    def fwdprop(cls, a, b, val):
        """
        Propagation for forward mode. Suppose current operation : c = a op b is one step of f(x), by chain rule,
        we have: dc/dx = dc/da * da/dx + dc/db * db/dx, return dc/dx.
        Args:
            a: the first operand, a Var instance
            b: the second operand, a Var instance, could be None
            val: the numerical operation result

        Returns:
            a dict, the derivative of operation result instance
        """
        if cls.n_operands == 2:
            lda, ldb = cls.local_derivative(a.val, b.val, val,
                                            skip_lda=a.is_const,
                                            skip_ldb=b.is_const)

            return cls.chain_rule(lda, a.derivative, ldb, b.derivative)

        lda = cls.local_derivative(a.val, None, val)
        return cls.chain_rule(lda, a.derivative, None, None)

    @classmethod
    def backprop(cls, a, b, val, dfdc):
        """
        Propagation for reverse/mix mode. Suppose current operation : c = a op b is one step of f(x), by chain rule,
        we have: df/da = df/dc * dc/da, df/db = df/dc * dc/db.
        Args:
            a: the first operand, a Var instance
            b: the second operand, a Var instance, could be None
            val: the numerical operation result
            dfdc: the backprop gradient.

        Returns:
            None
        """
        if cls.n_operands == 2:
            lda, ldb = cls.local_derivative(a.val, b.val, val,
                                            skip_lda=a.is_const,
                                            skip_ldb=b.is_const)
            a._bpgrad.update(cls.chain_rule(lda, dfdc, None, None, False))
            b._bpgrad.update(cls.chain_rule(None, None, ldb, dfdc, False))

        else:
            lda = cls.local_derivative(a.val, None, val)
            a._bpgrad.update(cls.chain_rule(lda, dfdc, None, None, False))

    @classmethod
    def merge_fwd_backprop(cls, dcdxs, dfdc):
        """
        Merge derivatives from forward mode and reverse mode. Suppose current node is c, in mix mode. W.r.t x, we have
        dc/dx and df/dc, then PART of df/dx is df/dc * dc/dx.
        Args:
            dcdxs: a dict like {'x': dcdx, 'y': dcdy}
            dfdc: a dict like {f: dfdc}

        Returns:
            a dict like {'x': dfdc (part), 'y': dfdy (part)}
        """
        dfdxs = {}
        for wrt in dcdxs:
            dfdxs[wrt] = np.einsum("ijpq, klij->klpq", dcdxs[wrt], dfdc)

        return dfdxs

    @classmethod
    def binary_operation(cls, a, b):
        """
        A universal binary operation process. Newly defined operations (class) do not need to re-write it.
        Args:
            a: a Number of np.ndarray or `Var` instance, the first operand of the calculation
            b: a Number of np.ndarray or `Var` instance , the second operand of the calculation

        Returns:
            A `Var` instance whose `.val` is the numerical value of the operation and `.derivative` containing
            the derivative w.r.t. the involved variables.

        """

        if not isinstance(a, Var):
            a = Var(a)

        if not isinstance(b, Var):
            b = Var(b)

        # Stop numpy auto broadcasting but allow the operation between scalar and vector,
        # or the differentiation would be too complicated to deal with
        if np.ndim(a.val) > 0 and np.ndim(b.val) > 0 and cls.__name__ != "VarMatMul":
            assert a.val.shape == b.val.shape, f"Shapes mismatch: {a.val.shape} != {b.val.shape}"

        # S1: calculate numerical result
        val = cls.op(a.val, b.val)

        # S2: get mode of the result
        mode = cls.merge_modes(a.mode, b.mode)

        # Prepare params for constructing a Var instance to contain the operation result
        params = dict(derivative={},
                      _var_shapes=cls.merge_var_shapes(a._var_shapes, b._var_shapes),
                      mode=mode,
                      _context=[cls, [a, b]])

        # Reverse/mix mode vars will calculate derivative later (when .diff() is called)
        if mode not in (Mode.Forward, Mode.NONE):
            return Var(val, **params)

        params["derivative"] = cls.fwdprop(a, b, val)
        return Var(val, **params)

    @classmethod
    def unary_operation(cls, a):
        """
        A universal unary operation process. Newly defined operations (class) do not need to re-write it.

        Args:
            a: a Number of np.ndarray or `Var` instance, the first operand of the calculation

        Returns:
            A `Var` instance whose `.val` is the numerical value of the operation and `.derivative` containing
            the derivative w.r.t. the involved variables.
        """
        if not isinstance(a, Var):
            a = Var(a)

        # S1: calculate numerical result
        val = cls.op(a.val, None)
        # S2: inherit the mode for the result
        mode = a.mode
        # Prepare params for constructing a Var instance to contain the operation result
        params = dict(derivative={},
                      _var_shapes=cls.merge_var_shapes(a._var_shapes),
                      mode=mode,
                      _context=[cls, [a]])

        if mode not in (Mode.Forward, Mode.NONE):
            return Var(val, **params)

        params["derivative"] = cls.fwdprop(a, None, val)
        return Var(val, **params)


class VarNeg(Ops):
    """
    A class for negative constants. Gives the value and local derivative.
    This class inherits from the Ops Class.
    To use:
    >>> -Var(1, 'x')
    (<class 'pyautodiff.var.Var'> name: None val: -1, der: {'x': array([[[[-1.]]]])})
    """

    symbol = "-"

    @classmethod
    def op(cls, va, vb):
        return -va

    @classmethod
    def local_derivative(cls, va, vb, vc, **kwargs):
        return -1


class VarPos(Ops):
    """
    A class for positive constants. Gives the value and local derivative.        
    This class inherits from the Ops Class.
    >>> +Var(1, 'x')
    (<class 'pyautodiff.var.Var'> name: None val: 1, der: {'x': array([[[[1.]]]])})
    """

    symbol = "+"

    @classmethod
    def op(cls, va, vb):
        return va

    @classmethod
    def local_derivative(cls, va, vb, vc, **kwargs):
        return 1


class VarAbs(Ops):
    """
    A class for absolute values. Gives the value and local derivative. 
    This class inherits from the Ops Class.
    >>> abs(Var(1, 'x'))
    (<class 'pyautodiff.var.Var'> name: None val: 1, der: {'x': array([[[[1.]]]])})
    """

    @classmethod
    def op(cls, va, vb):
        return abs(va)

    @classmethod
    def local_derivative(cls, va, vb, vc, **kwargs):
        d = np.ones_like(va)
        d[va < 0] = -1

        try:
            # scalar
            return d.item()
        except:
            return d


class VarAdd(Ops):
    """
    A class for addition. Gives the value and local derivative.
    This class inherits from the Ops Class.
    >>> Var(1, 'x') + 1
    (<class 'pyautodiff.var.Var'> name: None val: 2, der: {'x': array([[[[1.]]]])})
    >>> Var(1, 'x') + Var(2, 'y')
    (<class 'pyautodiff.var.Var'> name: None val: 3, der: {'x': array([[[[1.]]]]), 'y': array([[[[1.]]]])})
    """

    n_operands = 2
    symbol = "+"

    @classmethod
    def op(cls, va, vb):
        return va + vb

    @classmethod
    def local_derivative(cls, va, vb, vc, **kwargs):
        return 1, 1


class VarSub(Ops):
    """
    A class for addition. Gives the value and local derivative.
    This class inherits from the Ops Class.
    >>> Var(1, 'x') - 1
    (<class 'pyautodiff.var.Var'> name: None val: 0, der: {'x': array([[[[1.]]]])})
    >>> Var(1, 'x') - Var(2, 'y')
    (<class 'pyautodiff.var.Var'> name: None val: -1, der: {'x': array([[[[1.]]]]), 'y': array([[[[-1.]]]])})
    """

    n_operands = 2
    symbol = "-"

    @classmethod
    def op(cls, va, vb):
        return va - vb

    @classmethod
    def local_derivative(cls, va, vb, vc, **kwargs):
        return 1, -1


class VarMul(Ops):
    """
    A class for multiplication. Gives the value and local derivative.
    This class inherits from the Ops Class.
    >>> Var(1, 'x') * 2
    (<class 'pyautodiff.var.Var'> name: None val: 2, der: {'x': array([[[[2.]]]])})
    >>> Var(1, 'x') * Var(2, 'y')
    (<class 'pyautodiff.var.Var'> name: None val: 2, der: {'x': array([[[[2.]]]]), 'y': array([[[[1.]]]])})
    """

    n_operands = 2
    symbol = "*"

    @classmethod
    def op(cls, va, vb):
        return va * vb

    @classmethod
    def local_derivative(cls, va, vb, vc, **kwargs):
        return vb, va


class VarTrueDiv(Ops):
    """
    A class for division. Gives the value and local derivative.     
    This class inherits from the Ops Class.
    >>> Var(4, 'x') / 2
    (<class 'pyautodiff.var.Var'> name: None val: 2.0, der: {'x': array([[[[0.5]]]])})
    >>> Var(4, 'x') / Var(2, 'y')
    (<class 'pyautodiff.var.Var'> name: None val: 2.0, der: {'x': array([[[[0.5]]]]), 'y': array([[[[-1.]]]])})
    """

    n_operands = 2
    symbol = "/"

    @classmethod
    def op(cls, va, vb):
        return va / vb

    @classmethod
    def local_derivative(cls, va, vb, vc, skip_lda=False, skip_ldb=False):
        if skip_ldb:
            return 1 / vb, 0

        if skip_lda:
            return 0, -vc / vb

        return 1 / vb, -vc / vb


class VarPow(Ops):
    """
    A class for power operation. Gives the value and local derivative.    
    This class inherits from the Ops Class.
    >>> Var(2, 'x') ** 2
    (<class 'pyautodiff.var.Var'> name: None val: 4, der: {'x': array([[[[4.]]]])})
    >>> Var(4, 'x') ** Var(2, 'y')
    (<class 'pyautodiff.var.Var'> name: None val: 16, der: {'x': array([[[[8.]]]]), 'y': array([[[[22.18070978]]]])})
    """

    n_operands = 2
    symbol = "power"

    @classmethod
    def op(cls, va, vb):
        return va ** vb

    @classmethod
    def local_derivative(cls, va, vb, vc, skip_lda=False, skip_ldb=False):
        """ derivative of w.r.t vb and vb: b * a^(b-1)*a', a^b*ln(a)*b' """

        if skip_ldb:
            return vb * (va ** (vb - 1)), 0

        if skip_lda:
            return 0, np.log(va) * vc

        return vb * (va ** (vb - 1)), np.log(va) * vc


class VarExp(Ops):
    """
    A class for exponential operation. Gives the value and local derivative.
    This class inherits from the Ops Class.
    >>> exp(Var(0, 'x'))
    (<class 'pyautodiff.var.Var'> name: None val: 1.0, der: {'x': array([[[[1.]]]])})
    """

    @classmethod
    def op(cls, va, vb):
        return np.exp(va)

    @classmethod
    def local_derivative(cls, va, vb, vc, **kwargs):
        """ c = e^a --> c' = e^a"""
        return vc


class VarLog(Ops):
    """
    A class for logarithm. Gives the value and local derivative.
    This class inherits from the Ops Class.
    >>> log(Var(1, 'x'))
    (<class 'pyautodiff.var.Var'> name: None val: 0.0, der: {'x': array([[[[1.]]]])})
    """

    n_operands = 2

    @classmethod
    def op(cls, va, vb):
        """ log_vb(va) """
        return np.log(va) / np.log(vb)
        # return np.log(va)

    @classmethod
    def local_derivative(cls, va, vb, vc, skip_lda=False, skip_ldb=False):
        """ c'=a'/(a*ln(b)), c' = -(b'*log(a))/(b*log^2(b))) """
        inv_log_vb = 1 / np.log(vb)

        if skip_ldb:
            return 1 / va * inv_log_vb, 0

        if skip_lda:
            return 0, -vc * inv_log_vb / vb

        return 1 / va * inv_log_vb, -vc * inv_log_vb / vb

    @classmethod
    def binary_operation_with_base(clf, a, base=math.e):
        """ Wrap function to explicitly specify base """
        return clf.binary_operation(a, base)


class VarLogistic(Ops):
    """
    Logistic function: f(x) = 1 / (1 + exp(x))
    >>> sigmoid((Var(0, 'x')))
    (<class 'pyautodiff.var.Var'> name: None val: 0.5, der: {'x': array([[[[0.25]]]])})
    """

    @classmethod
    def op(cls, va, vb):
        return 1 / (1 + np.exp(-va))

    @classmethod
    def local_derivative(cls, va, vb, vc, skip_lda=False, skip_ldb=False):
        return vc * (1 - vc)


class VarMatMul(Ops):
    """
    Matrix multiplication.
    >>> (Var(np.array([[1],[2]]), 'x') @ Var(np.array([[0,1]]), 'y')).val.tolist()
    [[0, 1], [0, 2]]
    """
    n_operands = 2
    symbol = "@"

    @classmethod
    def op(cls, va, vb):
        return va @ vb

    @classmethod
    def local_derivative(cls, va, vb, vc, skip_lda=False, skip_ldb=False):
        return vb, va

    @classmethod
    def chain_rule(cls, lda, da, ldb, db, forward_mode=True):
        """
        Apply chain rule in forward mode for Matmul: c = a@b
        Args:
            lda: A Number or np.ndarray, represents the local derivative: dc/da (b.val in this case)
            da: a dict stores the derivative of a.
                For example, {'x': da/dx, 'y': da/dy}, where `x`,'y' is the involved variables.
            ldb: A Number or np.ndarray, represents the local derivative: dc/db (a.val in this case)
            db: a dict stores the derivative of b.
                For example, {'x': db/dx, 'y': db/dy}, where `x`,'y' is the involved variables.

        Returns:
            A dict stores the derivative of c by applying the chain rule. For example,
            {'x': dc/dx, 'y': dc/dy} where `x`,'y' is the involved variables.

        """

        def _apply(d, ld, s):
            if d is None:
                return

            for wrt in d:
                dc[wrt] += np.einsum(s, d[wrt], ld)

        dc = defaultdict(int)
        _apply(da, lda, "pqkl,qr->prkl" if forward_mode else "mnpr,qr->mnpq")
        _apply(db, ldb, "qrkl,pq->prkl" if forward_mode else "mnpr,pq->mnqr")

        return dict(dc)


class VarTranspose(Ops):
    """
    Transpose matrix.
    >>> (Var(np.array([[1,2]]), 'x').T).val.tolist()
    [[1], [2]]
    """
    symbol = ".T"

    @classmethod
    def op(cls, va, vb):
        return np.transpose(va)

    @classmethod
    def fwdprop(cls, a, b, val):
        der = {}
        for wrt in a.derivative:
            der[wrt] = np.einsum('ijkl->jikl', a.derivative[wrt])

        return der

    @classmethod
    def backprop(cls, a, b, val, dfdc):
        bp = {}
        for wrt in dfdc:
            bp[wrt] = np.einsum('ijkl->ijlk', dfdc[wrt])

        a._bpgrad.update(bp)

    @classmethod
    def unary_operation(cls, a):
        """
        A universal unary operation process. Newly defined operations (class) do not need to re-write it.

        Args:
            a: a Number of np.ndarray or `Var` instance, the first operand of the calculation

        Returns:
            A `Var` instance whose `.val` is the numerical value of the operation and `.derivative` containing
            the derivative w.r.t. the involved variables.
        """
        if not isinstance(a, Var):
            a = Var(a)

        val = cls.op(a.val, None)

        mode = a.mode
        params = dict(derivative={},
                      _var_shapes=cls.merge_var_shapes(a._var_shapes),
                      mode=mode,
                      _context=[cls, [a]])

        if mode not in (Mode.Forward, Mode.NONE):
            return Var(val, **params)

        params["derivative"] = cls.fwdprop(a, None, val)
        return Var(val, **params)


Var.__neg__ = _dunder_wrapper(VarNeg.unary_operation, True)
Var.__pos__ = _dunder_wrapper(VarPos.unary_operation, True)
Var.__abs__ = _dunder_wrapper(VarAbs.unary_operation, True)

# +=, -=, *=, /= are auto enabled.

Var.__radd__ = Var.__add__ = _dunder_wrapper(VarAdd.binary_operation)
Var.__sub__ = _dunder_wrapper(VarSub.binary_operation)
Var.__rsub__ = lambda a, b: VarSub.binary_operation(b, a)
Var.__rmul__ = Var.__mul__ = _dunder_wrapper(VarMul.binary_operation)
Var.__truediv__ = _dunder_wrapper(VarTrueDiv.binary_operation)
Var.__rtruediv__ = lambda a, b: VarTrueDiv.binary_operation(b, a)

Var.__pow__ = _dunder_wrapper(VarPow.binary_operation)
Var.__rpow__ = lambda a, b: VarPow.binary_operation(b, a)
pow = VarPow.binary_operation

# TODO: Fan
Var.__matmul__ = _dunder_wrapper(VarMatMul.binary_operation)
Var.transpose = transpose = _dunder_wrapper(VarTranspose.unary_operation, True)

exp = VarExp.unary_operation  # enable exp(x)
Var.exp = _dunder_wrapper(exp, True)  # enable x.exp()

log = VarLog.binary_operation_with_base
Var.log = _dunder_wrapper(log)

logistic = sigmoid = VarLogistic.unary_operation
Var.sigmoid = _dunder_wrapper(sigmoid, True)

sqrt = lambda x: x ** 0.5
Var.sqrt = sqrt



import numpy as np
from matplotlib import pyplot as plt

from pyautodiff import *


def Newton_Raphson_method(fn, xk, stepsize_thresh=1e-6, max_iters=1000, success_tolerance=1e-6, debug=False):
    """
        Newton's method to find root.
        Args:
            fn: funtion
            xk: initial guess
            stepsize_thresh: If ||x_{k+1} - x_{k}|| <= thresh, return
            max_iters: If #iters > max_iters, return
            success_tolerance: The absolute tolerance for fn(root)
            debug: Defaults to False. If True, print info for every iteration
        Returns:
            A dict
        """

    f = None
    is_scalar = (np.ndim(xk) == 0)

    checker = abs if is_scalar else np.linalg.norm
    solver = (lambda x, y: y / x) if is_scalar else np.linalg.solve

    offset = 1

    for k in range(max_iters):
        f = fn(Var(xk, "x"))  # This is a Var instance!! Access val and der by .val and .diff() respectively
        delta_x = solver(f.diff(), -f.val)

        if checker(delta_x) < stepsize_thresh:
            offset = 0
            break

        if debug:
            print(f"k={k}\tx={np.round(xk, 2)}\tf(x)={np.round(f.val)}\tf'(x)={np.round(f.diff())}")

        xk = xk + delta_x

    return {
        "succeed": np.allclose(f.val, 0, atol=success_tolerance),
        "iter": k + offset,
        "x": xk,
        "f(x)": f.val,
        "f\'(x)": f.diff()
    }


def cal_val_der(fn, xs):
    vals = []
    ders = []

    for x in xs:
        try:
            if not isinstance(x, (Var, VarAutoName)):
                y = fn(VarAutoName(x))
            else:
                y = fn(x)
        except:
            y = Var(0)
        finally:
            vals.append(y.val)
            ders.append(y.diff())

    return vals, ders


def draw_scalar(fn, roots, plt_range=[0, 10]):
    x = np.linspace(plt_range[0], plt_range[1], 1000).tolist()
    y, d = cal_val_der(fn, x)

    fig, ax = plt.subplots()
    ax.plot(x, y, label='val')
    ax.plot(x, d, label='der')
    ax.scatter(roots, cal_val_der(fn, roots)[0], label="root")
    ax.grid(True, which='both')

    ax.axhline(y=0, color='k')
    ax.axvline(x=0, color='k')

    plt.title("Use 0 to fill in +-inf")
    plt.legend()
    plt.show()


if __name__ == '__main__':
    print("====Scalar demo====")
    f = lambda x: x ** (-x) - log(x)

    rtn = Newton_Raphson_method(f, 1, debug=True)

    if rtn['succeed']:
        root = rtn["x"]
        print(f"Find a root={np.round(root, 4)}")
        draw_scalar(f, [root], plt_range=[0.1, root + 0.5])
    else:
        print(f"Failed. Try another x0 or larger max_iters!")
        print(rtn)
        draw_scalar(f, [], plt_range=[1, 5])

    print("====Vector demo====")
    A = Var(np.array([[1, 2], [3, 4]]))
    g = lambda x: A @ x - sin(exp(x))
    n_roots = 0
    for x0 in [[1, -1], [1, 1], [0, 0]]:
        x0 = np.array(x0).reshape(-1, 1)
        rtn = Newton_Raphson_method(g, x0, debug=False)
        if rtn["succeed"]:
            n_roots += 1
            root = rtn["x"]
            print(f"Find #{n_roots} root={np.round(root, 2).tolist()}")
        else:
            print(f"Failed. Try another x0 or larger max_iters!")



import numpy as np

from pyautodiff import Var, Ops
from pyautodiff.ops import _dunder_wrapper


class VarSin(Ops):
    """
        A class for trigonometric sine operation. Gives the value and local derivative.

        This class inherits from the Ops Class.

    >>> sin(Var(0, 'x'))
    (<class 'pyautodiff.var.Var'> name: None val: 0.0, der: {'x': array([[[[1.]]]])})
    """

    @classmethod
    def op(cls, va, vb):
        return np.sin(va)

    @classmethod
    def local_derivative(cls, va, vb, vc, **kwargs):
        return np.cos(va)


class VarCos(Ops):
    """
        A class for trigonometric cosine operation. Gives the value and local derivative.
        
        This class inherits from the Ops Class.

    >>> cos(Var(0, 'x'))
    (<class 'pyautodiff.var.Var'> name: None val: 1.0, der: {'x': array([[[[0.]]]])})
    """

    @classmethod
    def op(cls, va, vb):
        return np.cos(va)

    @classmethod
    def local_derivative(cls, va, vb, vc, **kwargs):
        return -np.sin(va)


class VarTan(Ops):
    """
        A class for trigonometric tangent operation. Gives the value and local derivative.
        
        This class inherits from the Ops Class.

    >>> tan(Var(0, 'x'))
    (<class 'pyautodiff.var.Var'> name: None val: 0.0, der: {'x': array([[[[1.]]]])})

    """

    @classmethod
    def op(cls, va, vb):
        return np.tan(va)

    @classmethod
    def local_derivative(cls, va, vb, vc, **kwargs):
        return (1 / np.cos(va)) ** 2


class VarArcSin(Ops):
    """
        A class for trigonometric arcsine operation. Gives the value and local derivative.
        
        This class inherits from the Ops Class.

    >>> arcsin(Var(0, 'x'))
    (<class 'pyautodiff.var.Var'> name: None val: 0.0, der: {'x': array([[[[1.]]]])})

    """

    @classmethod
    def op(cls, va, vb):
        return np.arcsin(va)

    @classmethod
    def local_derivative(cls, va, vb, vc, **kwargs):
        return 1 / np.sqrt(1 - va ** 2)


class VarArcCos(Ops):
    """
        A class for trigonometric arccosine operation. Gives the value and local derivative.
        
        This class inherits from the Ops Class.

    >>> arccos(Var(0, 'x'))
    (<class 'pyautodiff.var.Var'> name: None val: 1.5707963267948966, der: {'x': array([[[[-1.]]]])})
    """

    @classmethod
    def op(cls, va, vb):
        return np.arccos(va)

    @classmethod
    def local_derivative(cls, va, vb, vc, **kwargs):
        return -1 / np.sqrt(1 - va ** 2)


class VarArcTan(Ops):
    """
        A class for trigonometric arctangent operation. Gives the value and local derivative.
        
        This class inherits from the Ops Class.

    >>> arctan(Var(0, 'x'))
    (<class 'pyautodiff.var.Var'> name: None val: 0.0, der: {'x': array([[[[1.]]]])})
    """

    @classmethod
    def op(cls, va, vb):
        return np.arctan(va)

    @classmethod
    def local_derivative(cls, va, vb, vc, **kwargs):
        return 1 / (1 + va ** 2)


class VarSinH(Ops):
    """
        A class for trigonometric hyperbolic sine operation. Gives the value and local derivative.
        
        This class inherits from the Ops Class.

    >>> arcsinh(Var(0, 'x'))
    (<class 'pyautodiff.var.Var'> name: None val: 0.0, der: {'x': array([[[[1.]]]])})
    """

    @classmethod
    def op(cls, va, vb):
        return np.sinh(va)

    @classmethod
    def local_derivative(cls, va, vb, vc, **kwargs):
        """ derivative of sinh(x) = cosh(x)"""
        return np.cosh(va)


class VarCosH(Ops):
    """
        A class for trigonometric hyperbolic cosine operation. Gives the value and local derivative.
        
        This class inherits from the Ops Class.

    >>> arccosh(Var(2, 'x'))
    (<class 'pyautodiff.var.Var'> name: None val: 1.3169578969248166, der: {'x': array([[[[0.57735027]]]])})
    """

    @classmethod
    def op(cls, va, vb):
        return np.cosh(va)

    @classmethod
    def local_derivative(cls, va, vb, vc, **kwargs):
        """ derivative of cosh(x) = sinh(x)"""
        return np.sinh(va)


class VarTanH(Ops):
    """
        A class for trigonometric hyperbolic tangent operation. Gives the value and local derivative.
        
        This class inherits from the Ops Class.

    >>> tanh(Var(1, 'x'))
    (<class 'pyautodiff.var.Var'> name: None val: 0.7615941559557649, der: {'x': array([[[[0.41997434]]]])})
    """

    @classmethod
    def op(cls, va, vb):
        return np.tanh(va)

    @classmethod
    def local_derivative(cls, va, vb, vc, **kwargs):
        """ derivative of tanh(x) = 1 - tanh(x)^2

        Args:
            **kwargs:
        """
        return 1 - np.tanh(va) ** 2


class VarArcSinH(Ops):
    """
        A class for trigonometric hyperbolic arcsine operation. Gives the value and local derivative.
        
        This class inherits from the Ops Class.

    >>> arcsinh(Var(0, 'x'))
    (<class 'pyautodiff.var.Var'> name: None val: 0.0, der: {'x': array([[[[1.]]]])})

    """

    @classmethod
    def op(cls, va, vb):
        return np.arcsinh(va)

    @classmethod
    def local_derivative(cls, va, vb, vc):
        """ for all real va """
        return 1 / np.sqrt((va ** 2) + 1)


class VarArcCosH(Ops):
    """
        A class for trigonometric hyperbolic arccosine operation. Gives the value and local derivative.
        
        This class inherits from the Ops Class.

    >>> arccosh(Var(2, 'x'))
    (<class 'pyautodiff.var.Var'> name: None val: 1.3169578969248166, der: {'x': array([[[[0.57735027]]]])})

    """

    @classmethod
    def op(cls, va, vb):
        return np.arccosh(va)

    @classmethod
    def local_derivative(cls, va, vb, vc, **kwargs):
        """ for all real va>1

        Args:
            **kwargs:
        """
        assert (va > 1), "va should be greater than 1."
        return 1 / np.sqrt((va ** 2) - 1)


class VarArcTanH(Ops):
    """
        A class for trigonometric hyperbolic arctan operation. Gives the value and local derivative.
        
        This class inherits from the Ops Class.

    >>> arctanh(Var(0, 'x'))
    (<class 'pyautodiff.var.Var'> name: None val: 0.0, der: {'x': array([[[[1.]]]])})

    """

    @classmethod
    def op(cls, va, vb):
        """ the domain of arctanh is (-1, 1) """
        assert (np.abs(va) < 1), "The value inside arctanh should be between (-1, 1)."

        return np.arctanh(va)

    @classmethod
    def local_derivative(cls, va, vb, vc, **kwargs):
        """ derivative of arctanh(x) = 1 / (1-x^2)

        Args:
            **kwargs:
        """
        return 1 / (1 - va ** 2)


sin = VarSin.unary_operation  # enable sin(x)
Var.sin = _dunder_wrapper(sin, True)  # enable x.sin()

arcsin = VarArcSin.unary_operation
Var.arcsin = _dunder_wrapper(arcsin, True)

cos = VarCos.unary_operation
Var.cos = _dunder_wrapper(cos, True)

arccos = VarArcCos.unary_operation
Var.arccos = _dunder_wrapper(arccos, True)

tan = VarTan.unary_operation
Var.tan = _dunder_wrapper(tan, True)

arctan = VarArcTan.unary_operation
Var.arctan = _dunder_wrapper(arctan, True)

sinh = VarSinH.unary_operation
Var.sinh = _dunder_wrapper(sinh, True)

arcsinh = VarArcSinH.unary_operation
Var.arcsinh = _dunder_wrapper(arcsinh, True)

cosh = VarCosH.unary_operation
Var.cosh = _dunder_wrapper(cosh, True)

arccosh = VarArcCosH.unary_operation
Var.arccosh = _dunder_wrapper(arccosh, True)

tanh = VarTanH.unary_operation
Var.tanh = _dunder_wrapper(tanh, True)

arctanh = VarArcTanH.unary_operation
Var.arctanh = _dunder_wrapper(arctanh, True)



import time
from enum import Enum

import numpy as np
from graphviz import Digraph

from pyautodiff import Mode


class NodeType(Enum):
    var = 0
    const = 1
    operation = 2


class Viser(object):
    """
    A class to draw the computational graph and auto diff trace by graphviz.Digraph (directed graph).
    To distinguish the computational graph and the AD trace, two families of arrow will be used:
    1. For the computational graph, use full arrow. Black one represents the first operand, while one represents the
    second operand.
    2. For the AD trace, use half arrow. Black one represents the Forward mode and the white for the reverse/mix mode.
    If the arrow has the upper part, it indicates the var is the first operand; the lower part indicates the second
    operand.
    To use:
    Viser(Var(1, 'x') + Var(2, 'y'), draw_AD_trace=False)
    """

    def __init__(self, x, draw_AD_trace=False, horizontal=True):
        """

        Args:
            x: a Var instance containing the operation result
            draw_AD_trace: defaults to False, plot the computational graph. If True, draw the AD trace
            horizontal: defaults to True, extend the plot horizontally (lfet->right); If False, draw the plot ,
            vertically (top->down).
        """
        self.n_nodes = 1
        self.ad_trace = draw_AD_trace

        self.g = Digraph('Trace', format='png')
        if horizontal:
            self.g.attr(rankdir='LR')

        self.g.node("0", label="output")
        self._draw(x, "0")

    @staticmethod
    def _get_op_symbol(cls):
        """
        Return the symbol of operation to display on the plot. For example, symbol of VarAdd: "+".
        Args:
            cls: Operation class

        Returns:

        """
        if cls.symbol is None:
            return cls.__name__[3:]
        return cls.symbol

    def _get_unique_id(self):
        """
        Generate a unique id for node.
        Returns:
            a string for id
        """
        return f"{time.process_time()}_{self.n_nodes}"

    @staticmethod
    def _get_color(xtype):
        """
        Return the color for node by node(var) type.
        Args:
            xtype: node type

        Returns:
            a string for color
        """
        return {
            NodeType.var: None,
            NodeType.const: "darkseagreen2",
            NodeType.operation: "lavender",
        }[xtype]

    @staticmethod
    def _get_shape(xtype):
        """
        Return the shape for node by node(var) type.
        Args:
            xtype: node type

        Returns:
            a string for shape
        """
        return {
            NodeType.var: "oval",
            NodeType.const: "oval",
            NodeType.operation: "box",
        }[xtype]

    @staticmethod
    def _get_style(xtype):
        """
        Return the box style for node by node(var) type.
        Args:
            xtype: node type

        Returns:
            a string for box style
        """

        return {
            NodeType.var: None,
            NodeType.const: "filled",
            NodeType.operation: "filled",
        }[xtype]

    @staticmethod
    def _get_arrow(is_second_operand=False, ad=False, reverse_mode=False):
        """
        Return the arrow type for node by node(var) type. The arrow type see the docstring of class.
        Args:
            xtype: node type

        Returns:
            a string for arrow type
        """

        if ad:
            if reverse_mode:
                return "ornormal" if is_second_operand else "olnormal"

            return "rnormal" if is_second_operand else "lnormal"

        return "onormal" if is_second_operand else "normal"

    @staticmethod
    def _beatify_val(val):
        """Keep at most 3 digits for float"""
        return np.around(val, 3)

    def _draw(self, x, father, is_second_operand=False):
        """
        Draw the graph recursivley. The graph stores in self.g.
        Be careful, the direction of the arrow is determined by the propagation direction.
        Args:
            x: a var instance, a member of a composite operation
            father: x's "previous" node.
            is_second_operand: True/False

        Returns:
            None
        """
        try:
            cls, operands = x._context
            xid = self._get_unique_id()
            xlabel = self._get_op_symbol(cls)
            xtype = NodeType.operation
        except:
            operands = []

            if x.name is None:
                xid = self._get_unique_id()
                xlabel = f"{self._beatify_val(x.val)}"
                xtype = NodeType.const
            else:
                xid = xlabel = x.name
                xtype = NodeType.var

        self.g.node(xid, label=xlabel,
                    color=self._get_color(xtype),
                    shape=self._get_shape(xtype),
                    style=self._get_style(xtype))

        if father is not None:
            if self.ad_trace and x.mode != Mode.Forward:
                self.g.edge(father, xid, arrowhead=self._get_arrow(is_second_operand, True, True))
            else:
                self.g.edge(xid, father, arrowhead=self._get_arrow(is_second_operand, self.ad_trace, False))

        for i, t in enumerate(operands):
            self._draw(t, xid, i == 1)

    def show(self):
        """Show the plot. For IPython/jupyter notebook, call "self.g" directly"""
        self.g.view(cleanup=True, directory="/tmp")

    def save(self, path):
        """Pass in a string as path, save the plot to local"""
        self.g.render(path)



import setuptools

long_description = ("PyAutoDiff provides the ability to seamlessly calculate the gradient of a given function within "
                    "your Python code. By using automatic differentiation, this project addresses efficiency and "
                    "precision issues in symbolic and finite differentiation algorithms")

setuptools.setup(
    name="4dpyautodiff",
    version="1.0.0",
    author="cs107-JTGC",
    author_email="chenfanzhuang@g.harvard.edu",
    description="A simple library for auto differentiation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cs107-JTGC/cs107-FinalProject",
    packages=setuptools.find_packages(exclude=['docs', 'tests*', 'scripts']),
    install_requires=[
        "numpy",
        "graphviz"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



from .var import Var, VarAutoName, Mode
from .ops import Ops, exp, log, pow, sqrt, transpose, sigmoid, logistic
from .trig_ops import sin, cos, tan, sinh, cosh, tanh, arcsin, arccos, arctan, arcsinh, arccosh, arctanh
from .visualization import Viser